# p_ACT_meta_0.1.R 
rm(list=ls())
version=0.1
options(stringsAsFactors=FALSE)

###################################################################################################
# The following parameters will not usually need user modification, but can be tweaked if desired
###################################################################################################

# These parameters affect precision of p-values and speed of estimates.  The default settings request
# increased precision for values of p_ACT <.05, <.01, and <.00001 but may be altered as desired.
# Use 25000 for quick, reasonably precise p-values.  Use 25000000 for ultra-precise but slower p-values.
level1=25000
cutoff2=.05
level2=250000
cutoff3=.01
level3=2500000
cutoff4=.00001
level4=25000000


###############################################################
#Necessary libraries.
# If not already installed, the package 'mvtnorm' may be downloaded at
# http://cran.r-project.org/src/contrib/Descriptions/mvtnorm.html
# For > 100 tests, version 0.8-0 of mvtnorm (or higher) is required.
###############################################################
library(MASS)
library(mvtnorm)
library(corpcor)

###############################################################
# The remaining code should not need to be modified by the user
###############################################################

###############################################################
#Definition of function to compute p_ACT.  Function will be called repeatedly in code below.
###############################################################
# Use pmvnorm() function from mvtnorm package to compute multivariate normal probabilities
# For more information: 
# Genz A, Bretz F, Hothorn T (2006) mvtnorm: Multivariate normal and t distribution. R package version 0.8-0
# Genz A (1992) Numerical computation of multivariate normal probabilities.  J Comput Graph Stat 1:141-150
# Genz A (1993) Comparison of methods for the computation of multivariate normal probabilities.  Comput Sci Stat 25:400-405
###############################################################
         p_ACT<-function(lower,upper,v) {
            p_ACT=1-pmvnorm(lower=lower,upper=upper,sigma=v,maxpts=level1,abseps=.0000000000001)
            if (p_ACT<cutoff2) {
                p_ACT=1-pmvnorm(lower=lower,upper=upper,sigma=v,maxpts=level2,abseps=.0000000000001)
                if (p_ACT<cutoff3) {
                p_ACT=1-pmvnorm(lower=lower,upper=upper,sigma=v,maxpts=level3,abseps=.0000000000001)
                if (p_ACT<cutoff4) {
                    p_ACT=1-pmvnorm(lower=lower,upper=upper,sigma=v,maxpts=level4,abseps=.0000000000001)
                }
              }
            } 
            p_ACT
         }
###############################################################


###############################################################
# Main body of program
###############################################################

###############################################################
# Read in data
###############################################################
system('rm .RData',ignore.stderr=TRUE)
err=0
sink("p_ACT_meta.out")

# Read in zstats.csv
###############################################################
zstats=read.csv('zstats.csv',header=FALSE,row.names=1,as.is=TRUE)
L=dim(zstats)[1]
N_z=dim(zstats)[2]
if (N_z==1) {
   err=1
   cat("Error: only one column of Z-statistics!\n")
   cat("Need one column for each study in the meta-analysis\n")
}
snplist=row.names(zstats)
k=order(snplist)
zstats=zstats[k,]

# Read in weights.csv
###############################################################
temp=read.csv('weights.csv',header=FALSE,row.names=1,as.is=TRUE)
temp2=merge(snplist,temp,by.x=1,by.y=0)
weights=temp2[,-1]
row.names(weights)=temp2[,1]

if(dim(temp)[1]!=L|dim(temp)[2]!=N_z) {
   err=1
   cat("Error: zstats.csv and weights.csv should have same number of rows and columns. \n")
} else
if(dim(weights)[1]!=L) {
   err=1
   cat("Error: different SNP names in zstats.csv and weights.csv. \n")
} 

# Read in optional genelist.csv 
###############################################################
multigene=FALSE
if(file.access('genelist.csv')==0) {
   multigene=TRUE
   genelist=read.csv('genelist.csv',header=FALSE,as.is=TRUE)
   genelist=merge(snplist,genelist,by.x=1,by.y=2)
   if(dim(genelist)[1]!=L) {
      err=1
      cat("Error: genelist.csv and zstats.csv should include the same list of SNPs.\n")
   }

   k=order(genelist[,2],genelist[,1])
   z=zstats[k,]
   w=weights[k,]/sqrt(apply(weights[k,]^2,1,sum,na.rm=TRUE))
   gl=genelist[k,]
} else {
   z=zstats
   w=weights
}

# Read in two_stage.csv (only needed for two-stage design)
###############################################################
two_stage=FALSE
if (file.access('two_stage.csv')==0) {
   two_stage=TRUE
   two_stage_threshold = scan('two_stage.csv')
}

# Read in genotype files
###############################################################
j = 1
flag=0
geno=list(as.list(rep(1,L)))
if (!two_stage) {
   while(file.access(paste('geno',j,'.csv',sep=''))==0) {
       temp=read.csv(paste('geno',j,'.csv',sep=''),header=FALSE,row.names=1,as.is=TRUE)
       temp2=merge(snplist,temp,by.x=1,by.y=0,all.x=TRUE)
       temp3=merge(snplist,temp,by.x=1,by.y=0,all=TRUE)
       if (dim(temp3)[1]>L) {
          err=1 
          cat("Error: geno",j,".csv includes SNPs not listed in zstat.csv.  zstat.csv should include all SNPs tested.\n",sep='')
       }
       g=temp2[,-1]
       row.names(g)=temp2[,1]
       geno[[j]]=t(g[k,])
       j=j+1
   }
   N_studies=j-1
   if(N_studies!=N_z) {
      err=1
      cat("Error: number of columns of z-statistics (",N_z,") does not match number of geno files (",N_studies,").\n",sep='') 
   }
}
   
if (two_stage) {
    temp=read.csv('geno1.csv',header=FALSE,row.names=1,as.is=TRUE)
    temp2=merge(snplist,temp,by.x=1,by.y=0,all.x=TRUE)
    temp3=merge(snplist,temp,by.x=1,by.y=0,all=TRUE)
    if (dim(temp3)[1]>L) {
       err=1 
       cat("Error: geno1.csv includes SNPs not listed in zstat.csv.  zstat.csv should include all SNPs tested.\n",sep='')
    }
    g=temp2[,-1]
    row.names(g)=temp2[,1]
    geno[[1]]=t(g[k,])
    N_studies = 1
}

if (err!=0) {cat("Program terminated due to error(s) listed above\n")} 
sink()

if (err==0) {
###############################################################
# Recreate meta test-statistics
###############################################################
meta_z = apply(z*w,1,sum,na.rm=TRUE)
meta_max = max(abs(meta_z))
q_all=meta_max

###############################################################
# Fill in missing values (for correlation estimation only)
###############################################################
for (j in 1:N_studies) {
   for (i in 1:L) {
      mg=mean(geno[[j]][,i],na.rm=TRUE)
      geno[[j]][is.na(geno[[j]][,i]),i]=mg
   }
   geno[[j]][is.na(geno[[j]])]=0
}

###############################################################
# If two-stage design, compute z_ACT
###############################################################
if (two_stage) {
   
   # Check that two-stage design is consistent with threshold
   ###############################################################
   if (N_z>2) { check_2s = apply(!is.na(z[,-1]),1,sum) } else {check_2s = !is.na(z[,2])}
   lowest = min(abs(z[check_2s,1]))
   p_lowest=pnorm(-lowest)*2
   warn_2s=(p_lowest>two_stage_threshold)

   # Compute z_ACT
   ###############################################################
   cutoff = qnorm(1-two_stage_threshold/2)
   minicor = diag(2)
   minicor[1,2] = minicor[2,1] = mean(w[w[,1]<1,1])
   temp = 1 - (pnorm(cutoff)-pnorm(-cutoff)) - (pnorm(meta_max)-pnorm(-meta_max)) + 1 - p_ACT(c(-cutoff,-meta_max),c(cutoff,meta_max),minicor)
   zact = qnorm(1 - temp[1]/2)
   q_all=zact
}

###############################################################
# Compute p_ACT if only one gene
###############################################################
if (multigene==FALSE) {
   cormat=0
   if (two_stage) {
      cormat=cor.shrink(geno[[1]])
   } else {
      for (j in 1:N_studies) {
         cormat=cormat+(w[,j]%*%t(w[,j]))*cor.shrink(geno[[j]])
      }
   }
   d=dim(cormat)[1]
   p_ACT_experiment = p_ACT(-rep(q_all,d),rep(q_all,d),cormat) 
}

###############################################################
# Loop over genes, if multiple genes 
###############################################################
if (multigene==TRUE) {
   genenames=unique(genelist[,2])
   N_g=length(genenames)
   p_gene=p_exp=matrix(NA,N_g)
   for (i in 1:N_g) {
      snpset=(gl[,2]==genenames[i])
      q_gene = max(abs(meta_z[snpset]))
      # For genes with only 1 SNP
      if (sum(snpset)==1) { 
         if (!two_stage) {p_gene[i] = 2*pnorm(-q_gene)}
         p_exp[i] = 2*pnorm(-q_all)
      } else {
      # For genes with >1 SNPs
         cormat=0
         if (two_stage) {
            cormat=cor.shrink(geno[[1]][,snpset])
         } else {
            for (j in 1:N_studies) {
               cormat=cormat+(w[snpset,j]%*%t(w[snpset,j]))*cor.shrink(geno[[j]][,snpset])
            }
         }
         d=dim(cormat)[1]
         if (!two_stage) {p_gene[i] = p_ACT(-rep(q_gene,d),rep(q_gene,d),cormat)}
         if (q_gene!=q_all) {
            p_exp[i] = p_ACT(-rep(q_all,d),rep(q_all,d),cormat)
         } else {p_exp[i] = p_gene[i]}
      }
   }
   row.names(p_gene)=genenames
   p_ACT_experiment = 1 - prod(1 - p_exp)
}
sink("p_ACT_meta.out",append=TRUE)
if (!two_stage) {cat("\n\nMeta-analysis of",L,"SNPs across",N_studies,"studies\n")}
if (two_stage) {cat("\n\nTwo-stage analysis of",L,"SNPs\n")}
cat("############################################\n\n")
if (two_stage & warn_2s) {
   cat("Warning: based on the information in zstats.csv, it appears that SNPs with p>",two_stage_threshold,"were followed up in the second stage. Specifically, a SNP with |Z|=",lowest,"would have p-value",p_lowest,"and would not pass the specified threshold of p <",two_stage_threshold," for inclusion in stage 2.  If the threshold specified in two_stage.csv is incorrect, the following p-value is likely to be invalid: ")
}
cat("Experiment-wide p_ACT =",format(p_ACT_experiment,digits=4),"\n\n")
cat("############################################\n")

if (multigene & !two_stage) {
   cat("\n\nAdjusted p-values for each gene:\n")
   write.table(format(p_gene,digits=4),quote=FALSE,col.names=FALSE)
}

cat("\n\nMeta-analysis Z-statistics for individual SNPs\n")
cat("(based on information provided in zstats.csv and weights.csv)\n")
write.table(format(meta_z,digits=3),quote=FALSE,col.names=FALSE)
cat("\n\n\n")
}
